import asyncio
from concurrent.futures.thread import ThreadPoolExecutor
from enum import Enum
from typing import cast

import pytest
from mock import MagicMock

from opentrons.drivers.absorbance_reader import (
    AbsorbanceHidInterface,
    AbsorbanceReaderDriver,
)
from opentrons.drivers.absorbance_reader.async_byonoy import AsyncByonoy
from opentrons.drivers.types import ABSMeasurementMode, AbsorbanceReaderLidStatus


@pytest.fixture
def mock_interface() -> MagicMock:
    return MagicMock(spec=AbsorbanceHidInterface)


@pytest.fixture
def mock_device() -> MagicMock:
    return MagicMock(spec=AbsorbanceHidInterface.Device)


class MockErrorCode(Enum):
    NO_ERROR = "no_error"
    ERROR = "error"


@pytest.fixture
async def mock_async_byonoy(
    mock_interface: MagicMock, mock_device: MagicMock
) -> AsyncByonoy:
    loop = asyncio.get_running_loop()
    return AsyncByonoy(
        mock_interface, mock_device, ThreadPoolExecutor(max_workers=1), loop
    )


@pytest.fixture
async def driver(mock_async_byonoy: AsyncByonoy) -> AbsorbanceReaderDriver:
    return AbsorbanceReaderDriver(mock_async_byonoy)


@pytest.fixture
async def connected_driver(driver: AbsorbanceReaderDriver) -> AbsorbanceReaderDriver:
    driver._connection._device_handle = 1
    return driver


async def test_driver_connect_disconnect(
    mock_interface: MagicMock,
    driver: AbsorbanceReaderDriver,
) -> None:
    mock_interface.open_device.return_value = (
        MockErrorCode.NO_ERROR,
        1,
    )

    assert not await driver.is_connected()
    await driver.connect()

    mock_interface.open_device.assert_called_once()
    assert await driver.is_connected()
    assert driver._connection._verify_device_handle()
    assert driver._connection._device_handle == 1

    mock_interface.free_device.return_value = MockErrorCode.NO_ERROR
    await driver.disconnect()

    assert not await driver.is_connected()
    assert driver._connection._device_handle is None


async def test_driver_get_device_info(
    mock_interface: MagicMock,
    connected_driver: AbsorbanceReaderDriver,
) -> None:
    DEVICE_INFO = MagicMock(AbsorbanceHidInterface.DeviceInfo)
    DEVICE_INFO.ref_no = "DE MAA 001"
    DEVICE_INFO.sn = "BYOMAA00013"
    DEVICE_INFO.version = "Absorbance V1.0.2 2024-04-18"

    mock_interface.get_device_information.return_value = (
        MockErrorCode.NO_ERROR,
        DEVICE_INFO,
    )

    info = await connected_driver.get_device_info()

    assert info == {"serial": "BYOMAA00013", "model": "ABS96", "version": "v1.0.2"}
    mock_interface.get_device_information.assert_called_once()
    mock_interface.reset_mock()

    # Test Device info with updated serial format
    DEVICE_INFO.sn = "OPTMAA00034"
    DEVICE_INFO.version = "Absorbance V1.0.2 2024-04-18"

    mock_interface.get_device_information.return_value = (
        MockErrorCode.NO_ERROR,
        DEVICE_INFO,
    )

    info = await connected_driver.get_device_info()

    assert info == {"serial": "OPTMAA00034", "model": "ABS96", "version": "v1.0.2"}
    mock_interface.get_device_information.assert_called_once()
    mock_interface.reset_mock()

    # Test Device info with invalid serial format
    DEVICE_INFO.sn = "YRFGHVMAA00034"
    DEVICE_INFO.version = "Absorbance V1.0.2 2024-04-18"

    mock_interface.get_device_information.return_value = (
        MockErrorCode.NO_ERROR,
        DEVICE_INFO,
    )

    info = await connected_driver.get_device_info()

    assert info == {"serial": "OPTMAA00000", "model": "ABS96", "version": "v1.0.2"}
    mock_interface.get_device_information.assert_called_once()
    mock_interface.reset_mock()

    # Test Device info with updated version format
    DEVICE_INFO.sn = "OPTMAA00034"
    DEVICE_INFO.version = "8"

    mock_interface.get_device_information.return_value = (
        MockErrorCode.NO_ERROR,
        DEVICE_INFO,
    )

    info = await connected_driver.get_device_info()

    assert info == {"serial": "OPTMAA00034", "model": "ABS96", "version": "8"}
    mock_interface.get_device_information.assert_called_once()
    mock_interface.reset_mock()

    # Test Device info with invalid version format
    DEVICE_INFO.sn = "OPTMAA00034"
    DEVICE_INFO.version = "asd"

    mock_interface.get_device_information.return_value = (
        MockErrorCode.NO_ERROR,
        DEVICE_INFO,
    )

    info = await connected_driver.get_device_info()

    assert info == {"serial": "OPTMAA00034", "model": "ABS96", "version": "v0"}
    mock_interface.get_device_information.assert_called_once()
    mock_interface.reset_mock()


@pytest.mark.parametrize(
    "parts_aligned, module_status",
    [(True, AbsorbanceReaderLidStatus.ON), (False, AbsorbanceReaderLidStatus.OFF)],
)
async def test_driver_get_lid_status(
    mock_interface: MagicMock,
    connected_driver: AbsorbanceReaderDriver,
    parts_aligned: bool,
    module_status: AbsorbanceReaderLidStatus,
) -> None:
    mock_interface.get_device_parts_aligned.return_value = (
        MockErrorCode.NO_ERROR,
        parts_aligned,
    )

    status = await connected_driver.get_lid_status()

    mock_interface.get_device_parts_aligned.assert_called_once()
    assert status == module_status


async def test_driver_get_supported_wavelengths(
    mock_interface: MagicMock,
    connected_driver: AbsorbanceReaderDriver,
) -> None:
    SUPPORTED_WAVELENGTHS = [450, 500]
    mock_interface.abs96_get_available_wavelengths.return_value = (
        MockErrorCode.NO_ERROR,
        SUPPORTED_WAVELENGTHS,
    )

    assert not connected_driver._connection._supported_wavelengths

    wavelengths = await connected_driver.get_available_wavelengths()

    mock_interface.abs96_get_available_wavelengths.assert_called_once()
    assert connected_driver._connection._supported_wavelengths == SUPPORTED_WAVELENGTHS
    assert wavelengths == SUPPORTED_WAVELENGTHS


async def test_driver_initialize_and_read_single(
    mock_interface: MagicMock,
    connected_driver: AbsorbanceReaderDriver,
) -> None:
    # set up mock interface
    connected_driver._connection._supported_wavelengths = [450, 500]
    mock_interface.abs96_initialize_single_measurement.return_value = (
        MockErrorCode.NO_ERROR
    )

    class MeasurementConfig(AbsorbanceHidInterface.SingleMeasurementConfig):
        def __init__(self) -> None:
            self.sample_wavelength = 0
            self.reference_wavelength = 0

    mock_interface.Abs96SingleMeasurementConfig = MeasurementConfig

    # current config should not have been setup yet
    assert not connected_driver._connection._current_config
    await connected_driver.initialize_measurement([450], mode=ABSMeasurementMode.SINGLE)

    conf = cast(
        AbsorbanceHidInterface.SingleMeasurementConfig,
        connected_driver._connection._current_config,
    )
    assert conf and conf.sample_wavelength == 450
    mock_interface.abs96_initialize_single_measurement.assert_called_once_with(1, conf)

    # setup up mock interface with a single reading
    MEASURE_RESULT = [[0.1] * 96]
    mock_interface.abs96_single_measure.return_value = (
        MockErrorCode.NO_ERROR,
        MEASURE_RESULT,
    )

    result = await connected_driver.get_measurement()
    mock_interface.abs96_single_measure.assert_called_once_with(1, conf)

    assert result == MEASURE_RESULT


async def test_driver_initialize_and_read_multi(
    mock_interface: MagicMock,
    connected_driver: AbsorbanceReaderDriver,
) -> None:
    # set up mock interface
    connected_driver._connection._supported_wavelengths = [450, 500, 600]
    mock_interface.abs96_initialize_multiple_measurement.return_value = (
        MockErrorCode.NO_ERROR
    )

    class MeasurementConfig(AbsorbanceHidInterface.MultiMeasurementConfig):
        def __init__(self) -> None:
            self.sample_wavelengths = [0]

    mock_interface.Abs96MultipleMeasurementConfig = MeasurementConfig

    # current config should not have been setup yet
    assert not connected_driver._connection._current_config
    await connected_driver.initialize_measurement(
        [450, 500, 600], mode=ABSMeasurementMode.MULTI
    )

    conf = cast(
        AbsorbanceHidInterface.MultiMeasurementConfig,
        connected_driver._connection._current_config,
    )
    assert conf and conf.sample_wavelengths == [450, 500, 600]
    mock_interface.abs96_initialize_multiple_measurement.assert_called_once_with(
        1, conf
    )

    # setup up mock interface with multiple readings
    MEASURE_RESULT = [[0.1] * 96, [0.2] * 96, [0.3] * 96]
    mock_interface.abs96_multiple_measure.return_value = (
        MockErrorCode.NO_ERROR,
        MEASURE_RESULT,
    )

    result = await connected_driver.get_measurement()
    mock_interface.abs96_multiple_measure.assert_called_once_with(1, conf)

    assert result == MEASURE_RESULT



"""Tests for opentrons.protocol_reader.file_hasher."""

from opentrons.protocol_reader import (
    BufferedFile,
    FileHasher,
)


async def test_hash_files_different_order() -> None:
    """It should return the same value when the same files are hashed in a different order."""
    file_1 = BufferedFile(
        name="some_protocol.py",
        contents=bytes("protocol content", encoding="utf-8"),
        path=None,
    )
    file_2 = BufferedFile(
        name="some_custom_labware.json",
        contents=bytes("labware content", encoding="utf-8"),
        path=None,
    )

    subject = FileHasher()
    result1 = await subject.hash([file_1, file_2])
    result2 = await subject.hash([file_2, file_1])

    assert result1 == result2


async def test_hash_files_same_contents_different_file_names() -> None:
    """It should return a different value when the same files are hashed and have different file names."""
    file_1 = BufferedFile(
        name="some_protocol.py",
        contents=bytes("some_content", encoding="utf-8"),
        path=None,
    )
    file_2 = BufferedFile(
        name="some_protocol.py.json",
        contents=bytes("some_content", encoding="utf-8"),
        path=None,
    )

    subject = FileHasher()
    result1 = await subject.hash([file_1])
    result2 = await subject.hash([file_2])

    assert result1 != result2


async def test_hash_files_swapped_names_and_content() -> None:
    """It should return a different value when name and contents are swapped."""
    file_1 = BufferedFile(
        name="some_protocol.json",
        contents=bytes("some_content", encoding="utf-8"),
        path=None,
    )
    file_2 = BufferedFile(
        name="some_protocol",
        contents=bytes(".jsonsome_content", encoding="utf-8"),
        path=None,
    )

    subject = FileHasher()
    result1 = await subject.hash([file_1])
    result2 = await subject.hash([file_2])

    assert result1 != result2



import inspect
from datetime import datetime
from typing import Union, cast

import pytest
from decoy import Decoy
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition2,
    Parameters2,
)
from opentrons_shared_data.labware.types import (
    LabwareDefinition2 as LabwareDef2Dict,
)
from opentrons_shared_data.labware.types import (
    LabwareUri,
)

from opentrons import calibration_storage
from opentrons import types as top_types
from opentrons.calibration_storage import helpers as calibration_storage_helpers
from opentrons.calibration_storage.ot2.models import v1 as v1_models
from opentrons.calibration_storage.ot3.models import v1 as ot3_models
from opentrons.config import feature_flags
from opentrons.hardware_control.instruments.ot2 import instrument_calibration as subject
from opentrons.hardware_control.instruments.ot3 import (
    instrument_calibration as subject_ot3,
)

SourceType = calibration_storage.types.SourceType
CalibrationStatus = calibration_storage.types.CalibrationStatus


@pytest.fixture(autouse=True)
def _use_mock_calibration_storage(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out the opentrons.calibration_storage module."""
    for name, func in inspect.getmembers(calibration_storage, inspect.isfunction):
        monkeypatch.setattr(calibration_storage, name, decoy.mock(func=func))
    for name, func in inspect.getmembers(calibration_storage.ot2, inspect.isfunction):
        monkeypatch.setattr(calibration_storage.ot2, name, decoy.mock(func=func))
    for name, func in inspect.getmembers(calibration_storage.ot3, inspect.isfunction):
        monkeypatch.setattr(calibration_storage.ot3, name, decoy.mock(func=func))

    for name, func in inspect.getmembers(
        calibration_storage_helpers, inspect.isfunction
    ):
        monkeypatch.setattr(calibration_storage_helpers, name, decoy.mock(func=func))

    monkeypatch.setattr(
        feature_flags,
        "enable_ot3_hardware_controller",
        decoy.mock(func=feature_flags.enable_ot3_hardware_controller),
    )
    monkeypatch.setattr(
        calibration_storage.ot3.gripper_offset,
        "get_gripper_jaw_width_data",
        decoy.mock(
            func=calibration_storage.ot3.gripper_offset.get_gripper_jaw_width_data
        ),
    )
    monkeypatch.setattr(
        calibration_storage.ot3.gripper_offset,
        "get_gripper_calibration_offset",
        decoy.mock(
            func=calibration_storage.ot3.gripper_offset.get_gripper_calibration_offset
        ),
    )
    monkeypatch.setattr(
        calibration_storage.ot3.gripper_offset,
        "save_gripper_jaw_width_data",
        decoy.mock(
            func=calibration_storage.ot3.gripper_offset.save_gripper_jaw_width_data
        ),
    )
    monkeypatch.setattr(
        calibration_storage.ot3.gripper_offset,
        "save_gripper_calibration",
        decoy.mock(
            func=calibration_storage.ot3.gripper_offset.save_gripper_calibration
        ),
    )


@pytest.fixture
def tip_rack_dict() -> LabwareDef2Dict:
    """Get a tip rack dictionary definition value object."""
    return cast(
        LabwareDef2Dict,
        {"namespace": "test", "version": 1, "parameters": {"loadName": "cool-labware"}},
    )


@pytest.fixture
def tip_rack_model() -> LabwareDefinition2:
    """Get a tip rack Pydantic model definition value object."""
    return LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        namespace="test",
        version=1,
        parameters=Parameters2.model_construct(  # type: ignore[call-arg]
            loadName="cool-labware",
            tipOverlap=None,  # add a None value to validate serialization to dictionary
        ),
    )


@pytest.mark.parametrize(
    "tip_rack_definition",
    [
        lazy_fixture("tip_rack_dict"),
        lazy_fixture("tip_rack_model"),
    ],
)
def test_load_tip_length(
    decoy: Decoy,
    tip_rack_dict: LabwareDef2Dict,
    tip_rack_definition: Union[LabwareDef2Dict, LabwareDefinition2],
) -> None:
    """Test that a tip length can be laoded for a pipette / tiprack combination."""
    tip_length_data = v1_models.TipLengthModel(
        tipLength=1.23,
        lastModified=datetime(year=2023, month=1, day=1),
        definitionHash="asdfghjk",
        source=subject.SourceType.factory,
        status=v1_models.CalibrationStatus(
            markedBad=True,
            source=subject.SourceType.user,
            markedAt=datetime(year=2023, month=2, day=2),
        ),
    )

    decoy.when(
        calibration_storage.ot2.load_tip_length_calibration(
            pip_id="abc123", definition=tip_rack_dict
        )
    ).then_return(tip_length_data)

    decoy.when(calibration_storage.helpers.hash_labware_def(tip_rack_dict)).then_return(
        "asdfghjk"
    )
    decoy.when(
        calibration_storage.helpers.uri_from_definition(tip_rack_dict)
    ).then_return(LabwareUri("def456"))

    result = subject.load_tip_length_for_pipette(
        pipette_id="abc123", tiprack=tip_rack_definition
    )

    assert result == subject.TipLengthCalibration(
        tip_length=1.23,
        source=subject.SourceType.factory,
        pipette="abc123",
        tiprack="asdfghjk",
        last_modified=datetime(year=2023, month=1, day=1),
        uri=LabwareUri("def456"),
        status=subject.CalibrationStatus(
            markedBad=True,
            source=subject.SourceType.user,
            markedAt=datetime(year=2023, month=2, day=2),
        ),
    )


@pytest.mark.parametrize(
    "left,right,ok",
    [
        # If either point is all 0 (uncalibrated) then the check should pass
        (top_types.Point(0, 0, 0), top_types.Point(0, 0, 2), True),
        (top_types.Point(0, 0, 2), top_types.Point(0, 0, 0), True),
        (top_types.Point(0, 0, 0), top_types.Point(0, 0, 0), True),
        # If both points are non-zero but all values are within the range the
        # check should pass
        (top_types.Point(0, 1.0, 1.5), top_types.Point(-1, 0, 0.2), True),
        # If both points are non-zero but at least one element is more than
        # the range different the test should fail
        (top_types.Point(0.1, -1, 4.3), top_types.Point(1.7, 0, 0.2), False),
        (top_types.Point(0.1, -3.2, 1.5), top_types.Point(0.6, 0.9, 1.3), False),
        (top_types.Point(0.1, -1, 1.5), top_types.Point(-0.2, -0.1, 6), False),
    ],
)
def test_instrument_consistency_check_ot3(
    left: top_types.Point, right: top_types.Point, ok: bool
) -> None:
    result = subject_ot3.check_instrument_offset_reasonability(left, right)
    if ok:
        assert result == []
    else:
        assert result[0].kind == "inconsistent-pipette-offset"
        assert result[0].offsets == {
            top_types.Mount.LEFT: left,
            top_types.Mount.RIGHT: right,
        }
        assert result[0].limit == 4.0


@pytest.mark.parametrize("gripper_id", ["phoney_baloney_gripper", None])
@pytest.mark.parametrize("loaded_encoder_val", [14.44444])
def test_load_gripper_jaw_width(
    gripper_id: str | None, decoy: Decoy, loaded_encoder_val: float
) -> None:
    default_return_val = subject_ot3.GripperJawWidthData(
        encoder_position_at_jaw_closed=None,
        source=SourceType.default,
        status=CalibrationStatus(),
    )
    _time = datetime.now()
    loaded_data = ot3_models.GripperJawWidthModel(
        encoderPositionAtJawClosed=loaded_encoder_val,
        source=SourceType.default,
        status=ot3_models.CalibrationStatus(),
        lastModified=_time,
    )
    nondefault_return_val = subject_ot3.GripperJawWidthData(
        encoder_position_at_jaw_closed=loaded_encoder_val,
        source=SourceType.default,
        status=CalibrationStatus(markedBad=False, source=None, markedAt=None),
    )
    decoy.when(feature_flags.enable_ot3_hardware_controller()).then_return(True)
    decoy.when(
        calibration_storage.ot3.gripper_offset.get_gripper_jaw_width_data(
            "phoney_baloney_gripper"
        )
    ).then_return(loaded_data)
    gripper_jaw_width = subject_ot3.load_gripper_jaw_width(gripper_id=gripper_id)
    if gripper_id is None:
        assert gripper_jaw_width == default_return_val
    else:
        if loaded_encoder_val is None:
            assert gripper_jaw_width == default_return_val
        else:
            assert gripper_jaw_width == nondefault_return_val


@pytest.mark.parametrize("gripper_id", ["phoney_baloney_gripper", None])
@pytest.mark.parametrize(
    "loaded_offset", [top_types.Point(0.1, -3.2, 1.5), top_types.Point(0, 0, 0)]
)
def test_load_gripper_calibration_offset(
    gripper_id: str | None, decoy: Decoy, loaded_offset: top_types.Point
) -> None:
    _time = datetime.now()
    default_return_val = subject_ot3.GripperCalibrationOffset(
        offset=top_types.Point(x=0.0, y=0.0, z=0.0),
        source=SourceType.default,
        status=CalibrationStatus(markedBad=False, source=None, markedAt=None),
        last_modified=None,
    )
    nondefault_return_val = subject_ot3.GripperCalibrationOffset(
        offset=loaded_offset,
        source=SourceType.default,
        status=CalibrationStatus(markedBad=False, source=None, markedAt=None),
        last_modified=_time,
    )

    loaded_data = ot3_models.InstrumentOffsetModel(
        offset=loaded_offset,
        lastModified=_time,
        source=SourceType.default,
        status=ot3_models.CalibrationStatus(),
    )
    decoy.when(feature_flags.enable_ot3_hardware_controller()).then_return(True)
    decoy.when(
        calibration_storage.ot3.gripper_offset.get_gripper_calibration_offset(
            "phoney_baloney_gripper"
        )
    ).then_return(loaded_data)
    gripper_offset = subject_ot3.load_gripper_calibration_offset(gripper_id=gripper_id)
    if gripper_id is None:
        assert gripper_offset == default_return_val
    else:
        if loaded_offset is None:
            assert gripper_offset == default_return_val
        else:
            assert gripper_offset == nondefault_return_val


@pytest.mark.parametrize("gripper_id", ["phoney_baloney_gripper", None])
@pytest.mark.parametrize("encoder_val", [14.44444])
def test_save_gripper_jaw_width_data(
    gripper_id: str | None, decoy: Decoy, encoder_val: float
) -> None:
    decoy.when(feature_flags.enable_ot3_hardware_controller()).then_return(True)
    subject_ot3.save_gripper_jaw_width_data(
        gripper_id=gripper_id, encoder_position_at_closed=encoder_val
    )
    decoy.verify(
        calibration_storage.ot3.gripper_offset.save_gripper_jaw_width_data(
            encoder_position_at_jaw_closed=encoder_val,
            gripper_id="phoney_baloney_gripper",
        ),
        times=1 if gripper_id is not None else 0,
    )


@pytest.mark.parametrize("gripper_id", ["phoney_baloney_gripper", None])
@pytest.mark.parametrize("delta", [top_types.Point(x=1, y=2, z=3)])
def test_save_gripper_calibration_offset(
    gripper_id: str | None, decoy: Decoy, delta: top_types.Point
) -> None:
    decoy.when(feature_flags.enable_ot3_hardware_controller()).then_return(True)
    subject_ot3.save_gripper_calibration_offset(gripper_id=gripper_id, delta=delta)
    decoy.verify(
        calibration_storage.ot3.gripper_offset.save_gripper_calibration(
            delta, "phoney_baloney_gripper"
        ),
        times=1 if gripper_id is not None else 0,
    )



"""File interaction resource provider."""

import csv
from dataclasses import dataclass
from datetime import datetime
from io import StringIO
from typing import Awaitable, Callable, Dict, List, Optional

from pydantic import BaseModel

from opentrons_shared_data.data_files import (
    DataFileInfo,
    MimeType,
    RunFileNameMetadata,
)

SPECIAL_CHARACTERS = {
    "#",
    "%",
    "&",
    "{",
    "}",
    "\\",
    "/",
    "<",
    ">",
    "*",
    "$",
    "!",
    "?",
    ".",
    "'",
    '"',
    ":",
    ";",
    "@",
    "`",
    "|",
}


@dataclass(frozen=True)
class FileNameCmdMetadata:
    """Command metadata associated with a specific data file."""

    command_id: str
    prev_command_id: str


@dataclass(frozen=True)
class ReadCmdFileNameMetadata(FileNameCmdMetadata):
    """Data from a plate reader `read` command used to build the finalized file name."""

    base_filename: str
    wavelength: int


@dataclass(frozen=True)
class ImageCaptureCmdFileNameMetadata(FileNameCmdMetadata):
    """Data from a camera capture command used to build the finalized file name."""

    step_number: int
    command_timestamp: datetime
    base_filename: Optional[str]


CommandFileNameMetadata = ReadCmdFileNameMetadata | ImageCaptureCmdFileNameMetadata


class FileData:
    """File data container for writing to a file."""

    data: bytes
    mime_type: MimeType
    run_metadata: RunFileNameMetadata
    command_metadata: CommandFileNameMetadata

    @staticmethod
    def build(
        data: bytes,
        mime_type: MimeType,
        run_metadata: RunFileNameMetadata,
        command_metadata: CommandFileNameMetadata,
    ) -> "FileData":
        """Build a generic file data class."""
        file_data = FileData()
        file_data.data = data
        file_data.mime_type = mime_type
        file_data.run_metadata = run_metadata
        file_data.command_metadata = command_metadata
        return file_data


class ReadData(BaseModel):
    """Read Data type containing the wavelength for a Plate Reader read alongside the Measurement Data of that read."""

    wavelength: int
    data: Dict[str, float]


class PlateReaderData(BaseModel):
    """Data from an Opentrons Plate Reader Read. Can be converted to CSV format."""

    read_results: List[ReadData]
    reference_wavelength: Optional[int] = None
    start_time: datetime
    finish_time: datetime
    serial_number: str

    def build_csv_bytes(self, measurement: ReadData) -> bytes:  # noqa: C901
        """Builds CSV data as bytes containing Plate Reader Measurements."""
        plate_alpharows = ["A", "B", "C", "D", "E", "F", "G", "H"]
        rows = []

        rows.append(["", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"])
        for i in range(8):
            row = [plate_alpharows[i]]
            for j in range(12):
                row.append(str(measurement.data[f"{plate_alpharows[i]}{j + 1}"]))
            rows.append(row)
        for i in range(3):
            rows.append([])
        rows.append(["", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"])
        for i in range(8):
            row = [plate_alpharows[i]]
            for j in range(12):
                row.append("")
            rows.append(row)
        for i in range(3):
            rows.append([])
        rows.append(
            [
                "",
                "ID",
                "Well",
                "Absorbance (OD)",
                "Mean Absorbance (OD)",
                "Absorbance %CV",
            ]
        )
        for i in range(3):
            rows.append([])
        rows.append(
            [
                "",
                "ID",
                "Well",
                "Absorbance (OD)",
                "Mean Absorbance (OD)",
                "Dilution Factor",
                "Absorbance %CV",
            ]
        )
        rows.append(["1", "Sample 1", "", "", "", "1", "", "", "", "", "", ""])
        for i in range(3):
            rows.append([])

        # end of file metadata
        rows.append(["Protocol"])
        rows.append(["Assay"])
        rows.append(["Sample Wavelength (nm)", str(measurement.wavelength)])
        if self.reference_wavelength is not None:
            rows.append(["Reference Wavelength (nm)", str(self.reference_wavelength)])
        rows.append(["Serial No.", self.serial_number])
        rows.append(
            ["Measurement started at", self.start_time.strftime("%m %d %H:%M:%S %Y")]
        )
        rows.append(
            ["Measurement finished at", self.finish_time.strftime("%m %d %H:%M:%S %Y")]
        )

        output = StringIO()
        writer = csv.writer(output, delimiter=",")
        writer.writerows(rows)
        csv_bytes = output.getvalue().encode("utf-8")

        return csv_bytes


class FileProvider:
    """Provider class to wrap file read write interactions to the data files directory in the engine."""

    def __init__(
        self,
        data_files_write_file_cb: Optional[
            Callable[[FileData], Awaitable[DataFileInfo]]
        ] = None,
    ) -> None:
        """Initialize the interface callbacks of the File Provider for data file handling within the Protocol Engine.

        Params:
            data_files_write_file_callback: Callback to write a file to the data files directory and add it to the database.
            data_files_filecount: Callback to check the amount of data files already present in the data files directory.
        """
        self._data_files_write_file_cb = data_files_write_file_cb
        self._run_metadata: RunFileNameMetadata | None = None

    async def write_file(
        self,
        data: bytes,
        mime_type: MimeType,
        command_metadata: CommandFileNameMetadata,
    ) -> DataFileInfo:
        """Writes arbitrary data to a file in the Data Files directory.

        Returns the `DataFileInfo` of the file created.

        Raises:
            Note that the callback may raise a StorageLimitReachedError.
        """
        if self._data_files_write_file_cb is not None:
            assert self._run_metadata is not None
            file_data = FileData.build(
                data=data,
                mime_type=mime_type,
                command_metadata=command_metadata,
                run_metadata=self._run_metadata,
            )

            return await self._data_files_write_file_cb(file_data)
        # If we are in an analysis or simulation state, return an empty `DataFileInfo`
        return DataFileInfo(
            id="",
            name="",
            file_hash="",
            created_at=datetime.now(),
            generated=True,
            stored=False,
            path="",
            mime_type=mime_type,
        )

    def set_run_metadata(self, metadata: RunFileNameMetadata) -> None:
        """Sets metadata specific to the run."""
        self._run_metadata = metadata

    def clear_run_metadata(self) -> None:
        """Clears metadata specific to the run."""
        self._run_metadata = None



"""Tests for `labware_offset_standardization`."""

from functools import lru_cache

import pytest

from opentrons_shared_data.deck import load
from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.robot.types import RobotType

from opentrons.protocol_engine import labware_offset_standardization as subject
from opentrons.protocol_engine.types import (
    LabwareOffsetCreate,
    LabwareOffsetCreateInternal,
    LabwareOffsetLocationSequence,
    LabwareOffsetVector,
    LegacyLabwareOffsetCreate,
    LegacyLabwareOffsetLocation,
    ModuleModel,
    OnAddressableAreaOffsetLocationSequenceComponent,
    OnLabwareOffsetLocationSequenceComponent,
    OnModuleOffsetLocationSequenceComponent,
)
from opentrons.types import DeckSlotName


@lru_cache
def load_from_robot_type(robot_type: RobotType) -> DeckDefinitionV5:
    """Get a deck from robot type."""
    if robot_type == "OT-3 Standard":
        return load("ot3_standard")
    else:
        return load("ot2_standard")


@pytest.mark.parametrize(
    ("location", "robot_type", "expected_modern_location", "expected_legacy_location"),
    [
        # Directly on a slot
        pytest.param(
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_5),
            "OT-2 Standard",
            [OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="5")],
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_5),
            id="direct-slot-ot2-native",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_5),
            "OT-3 Standard",
            [
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="C2"
                )
            ],
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_C2),
            id="direct-slot-flex-ot2",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-2 Standard",
            [OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="5")],
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_5),
            id="direct-slot-ot2-flex",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-3 Standard",
            [
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="C2"
                )
            ],
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_C2),
            id="direct-slot-flex-native",
        ),
        # On a module with no adapter
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-3 Standard",
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="module-flex-native",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-3 Standard",
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="module-flex-ot2",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-2 Standard",
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="module-ot2-flex",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-2 Standard",
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="module-ot2-native",
        ),
        # On a module with no adapter, and specificially the module is a Thermocycler.
        # The Thermocycler is special because the deck definition splits it into two
        # fixtures. :(
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_B1,
                moduleModel=ModuleModel.THERMOCYCLER_MODULE_V2,
            ),
            "OT-3 Standard",
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.THERMOCYCLER_MODULE_V2
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="thermocyclerModuleV2",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_B1,
                moduleModel=ModuleModel.THERMOCYCLER_MODULE_V2,
            ),
            id="tc-module-flex-native",
        ),
        # On a labware (or stack...) on a slot
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1, definitionUri="opentrons-test/foo/1"
            ),
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="D1"
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1, definitionUri="opentrons-test/foo/1"
            ),
            id="labware-slot-flex-native",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1, definitionUri="opentrons-test/foo/1"
            ),
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1, definitionUri="opentrons-test/foo/1"
            ),
            id="labware-slot-ot2-flex",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1, definitionUri="opentrons-test/foo/1"
            ),
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="D1"
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1, definitionUri="opentrons-test/foo/1"
            ),
            id="labware-slot-flex-ot2",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1, definitionUri="opentrons-test/foo/1"
            ),
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1, definitionUri="opentrons-test/foo/1"
            ),
            id="labware-slot-ot2-native",
        ),
        # On an adapter on a module
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-module-flex-native",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-module-ot2-flex",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-module-flex-ot2",
        ),
        pytest.param(
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-module-ot2-native",
        ),
    ],
)
def test_standardize_legacy_labware_offset(
    location: LegacyLabwareOffsetLocation,
    robot_type: RobotType,
    expected_modern_location: LabwareOffsetLocationSequence,
    expected_legacy_location: LegacyLabwareOffsetLocation,
) -> None:
    """It should convert deck slots in `LegacyLabwareOffsetCreate`s and go to the new format."""
    deck_def = load_from_robot_type(robot_type)
    original = LegacyLabwareOffsetCreate(
        definitionUri="opentrons-test/foo/1",
        location=location,
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    expected = LabwareOffsetCreateInternal(
        definitionUri="opentrons-test/foo/1",
        legacyLocation=expected_legacy_location,
        locationSequence=expected_modern_location,
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    assert (
        subject.standardize_labware_offset_create(original, robot_type, deck_def)
        == expected
    )


@pytest.mark.parametrize(
    ("location", "robot_type", "expected_modern_location", "expected_legacy_location"),
    [
        # Directly on a slot
        pytest.param(
            [OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="5")],
            "OT-2 Standard",
            [OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="5")],
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_5),
            id="slot-direct-ot2",
        ),
        pytest.param(
            [
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="C2"
                )
            ],
            "OT-3 Standard",
            [
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="C2"
                )
            ],
            LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_C2),
            id="slot-direct-flex",
        ),
        # On a module with no adapter
        pytest.param(
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            "OT-3 Standard",
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="module-slot-flex",
        ),
        pytest.param(
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            "OT-2 Standard",
            [
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="module-slot-ot2",
        ),
        # On a labware  on a slot
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="D1"
                ),
            ],
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="D1"
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1, definitionUri="opentrons-test/foo/1"
            ),
            id="labware-slot-flex",
        ),
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1, definitionUri="opentrons-test/foo/1"
            ),
            id="labware-slot-ot2",
        ),
        # On an adapter on a module
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="temperatureModuleV2D1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_D1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-module-flex",
        ),
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-slot-ot2",
        ),
        # On a stack of labware
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="A3",
                ),
            ],
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="A3",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_A3,
                definitionUri="opentrons-test/foo/1",
            ),
            id="labware-stack-flex",
        ),
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="2",
                ),
            ],
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="2",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_2,
                definitionUri="opentrons-test/foo/1",
            ),
            id="labware-stack-ot2",
        ),
        # On a stack of labware on a module
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3",
                ),
            ],
            "OT-2 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_3,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-stack-module-ot2",
        ),
        pytest.param(
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="A1",
                ),
            ],
            "OT-3 Standard",
            [
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/foo/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/bar/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/baz/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="A1",
                ),
            ],
            LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_A1,
                definitionUri="opentrons-test/foo/1",
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V2,
            ),
            id="labware-stack-module-flex",
        ),
    ],
)
def test_standardize_modern_labware_offset(
    location: LabwareOffsetLocationSequence,
    robot_type: RobotType,
    expected_modern_location: LabwareOffsetLocationSequence,
    expected_legacy_location: LegacyLabwareOffsetLocation,
) -> None:
    """It should convert deck slots in `LabwareOffsetCreate`s and fill in the old format."""
    deck_def = load_from_robot_type(robot_type)
    original = LabwareOffsetCreate(
        definitionUri="opentrons-test/foo/1",
        locationSequence=location,
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    expected = LabwareOffsetCreateInternal(
        definitionUri="opentrons-test/foo/1",
        legacyLocation=expected_legacy_location,
        locationSequence=expected_modern_location,
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    assert (
        subject.standardize_labware_offset_create(original, robot_type, deck_def)
        == expected
    )



"""opentrons.calibration_storage.helpers: various miscellaneous
functions

This module has functions that you can import to save robot or
labware calibration to its designated file location.
"""

import json
from dataclasses import asdict, is_dataclass
from hashlib import sha256
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union, cast

from . import types as local_types

if TYPE_CHECKING:
    from opentrons_shared_data.labware.types import LabwareDefinition
    from opentrons_shared_data.pipette.types import LabwareUri


def dict_filter_none(data: List[Tuple[str, Any]]) -> Dict[str, Any]:
    """
    Helper function to filter out None keys from a dataclass
    before saving to file.
    """
    return dict(item for item in data if item[1] is not None)


def convert_to_dict(obj: Any) -> Dict[str, Any]:
    # The correct way to type this is described here:
    # https://github.com/python/mypy/issues/6568
    # Unfortunately, since it's not currently supported I have an
    # assert check instead.
    assert is_dataclass(obj) and not isinstance(obj, type), (
        "This function is intended for dataclasses only"
    )
    return asdict(obj, dict_factory=dict_filter_none)


# TODO(mc, 2021-11-09): this hashing function may produce different hashes
# for semantically identical labware defs. For example, a value of `10` will
# produce a different hash than a value of `10.0`. Hashing should be produce
# identical hashes for `==` values. This could be done by running defs through
# Pydantic or similar rather via json.dumps
def hash_labware_def(labware_def: "LabwareDefinition") -> str:
    """
    Helper function to take in a labware definition and return
    a hashed string of key elements from the labware definition
    to make it a unique identifier.

    :param labware_def: Full labware definition
    :returns: sha256 string
    """
    # remove keys that do not affect run
    blocklist = ["metadata", "brand", "groups"]
    def_no_metadata = {k: v for k, v in labware_def.items() if k not in blocklist}
    sorted_def_str = json.dumps(def_no_metadata, sort_keys=True, separators=(",", ":"))

    return sha256(sorted_def_str.encode("utf-8")).hexdigest()


def details_from_uri(uri: str, delimiter: str = "/") -> local_types.UriDetails:
    """
    Unpack a labware URI to get the namespace, loadname and version
    """
    try:
        info = uri.split(delimiter)
        return local_types.UriDetails(
            namespace=info[0], load_name=info[1], version=int(info[2])
        )
    except IndexError:
        # Here we are assuming that the 'uri' passed in is actually
        # the loadname, though sometimes it may be an empty string.
        return local_types.UriDetails(namespace="", load_name=uri, version=1)


def uri_from_details(
    namespace: str, load_name: str, version: Union[str, int], delimiter: str = "/"
) -> "LabwareUri":
    """Build a labware URI from its details.

    A labware URI is a string that uniquely specifies a labware definition.

    :returns str: The URI.
    """
    return cast("LabwareUri", f"{namespace}{delimiter}{load_name}{delimiter}{version}")


def uri_from_definition(
    definition: "LabwareDefinition", delimiter: str = "/"
) -> "LabwareUri":
    """Build a labware URI from its definition.

    A labware URI is a string that uniquely specifies a labware definition.

    :returns str: The URI.
    """
    return uri_from_details(
        namespace=definition["namespace"],
        load_name=definition["parameters"]["loadName"],
        version=definition["version"],
        delimiter=delimiter,
    )



from typing import Dict

import pytest

from opentrons.hardware_control.types import Axis
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_api.labware import Labware, get_labware_definition
from opentrons.protocol_api.protocol_context import ProtocolContext
from opentrons.protocols.api_support.deck_type import STANDARD_OT2_DECK
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import (
    AxisMaxSpeeds,
    build_edges,
    find_value_for_api_version,
)
from opentrons.types import Location, Mount, Point


def test_max_speeds_userdict() -> None:
    defaults = AxisMaxSpeeds(robot_type="OT-2 Standard")
    assert defaults.data == {}
    assert dict(defaults) == {}

    with pytest.raises(KeyError):
        defaults["asdas"] = 2

    with pytest.raises(AssertionError):
        defaults["x"] = -1

    with pytest.raises(AssertionError):
        defaults["y"] = "ggg"

    with pytest.raises(KeyError):
        defaults["b"] = 2

    with pytest.raises(KeyError):
        defaults["c"] = 3

    defaults["x"] = 2
    defaults[Axis.A] = 20

    assert defaults["X"] == 2
    assert defaults["x"] == 2
    assert defaults[Axis.X] == 2

    assert defaults["A"] == 20
    assert defaults["a"] == 20
    assert defaults[Axis.A] == 20

    assert sorted(list(defaults.keys())) == sorted(["X", "A"])
    assert "X" in defaults.keys()

    del defaults["A"]
    assert "A" not in defaults

    defaults["x"] = None
    assert "x" not in defaults


def test_build_edges() -> None:
    lw_def = get_labware_definition("corning_96_wellplate_360ul_flat")
    assert lw_def["schemaVersion"] == 2  # LegacyLabwareCore expects this.
    test_lw = Labware(
        core=LegacyLabwareCore(lw_def, Location(Point(0, 0, 0), None)),
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )
    off = Point(0, 0, 1.0)
    deck = Deck(deck_type=STANDARD_OT2_DECK)
    old_correct_edges = [
        test_lw["A1"].from_center_cartesian(x=1.0, y=0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=-1.0, y=0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=1.0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=-1.0, z=1) + off,
    ]
    res = build_edges(test_lw["A1"], 1.0, Mount.RIGHT, deck, version=APIVersion(2, 2))
    assert res == old_correct_edges

    new_correct_edges = [
        test_lw["A1"].from_center_cartesian(x=1.0, y=0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=-1.0, y=0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=1.0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=-1.0, z=1) + off,
    ]
    res2 = build_edges(test_lw["A1"], 1.0, Mount.RIGHT, deck, version=APIVersion(2, 4))
    assert res2 == new_correct_edges


# TODO(mm, 2023-04-28): The build_edges() function is used both by ProtocolContexts
# that are backed by Protocol Engine, and those that aren't. But this test is only runnable
# with a non-Protocol-Engine ProtocolContext because it relies on the internal ctx._core.get_deck()
# property.
#
# Find a different way to test this so that both paths are covered.
@pytest.mark.apiv2_non_pe_only
def test_build_edges_left_pipette(ctx: ProtocolContext) -> None:
    test_lw = ctx.load_labware("corning_96_wellplate_360ul_flat", "2")
    test_lw2 = ctx.load_labware("corning_96_wellplate_360ul_flat", "6")
    mod = ctx.load_module("magnetic module", "3")
    mod.load_labware("corning_96_wellplate_360ul_flat")
    off = Point(0, 0, 1.0)
    left_pip_edges = [
        test_lw["A12"].from_center_cartesian(x=-1.0, y=0, z=1) + off,
        test_lw["A12"].from_center_cartesian(x=0, y=0, z=1) + off,
        test_lw["A12"].from_center_cartesian(x=0, y=1.0, z=1) + off,
        test_lw["A12"].from_center_cartesian(x=0, y=-1.0, z=1) + off,
    ]
    # Test that module in slot 3 results in modified edge list
    res = build_edges(
        test_lw["A12"],
        1.0,
        Mount.LEFT,
        ctx._core.get_deck(),  # type: ignore[attr-defined]
        version=APIVersion(2, 4),
    )
    assert res == left_pip_edges

    left_pip_edges = [
        test_lw2["A12"].from_center_cartesian(x=-1.0, y=0, z=1) + off,
        test_lw2["A12"].from_center_cartesian(x=0, y=0, z=1) + off,
        test_lw2["A12"].from_center_cartesian(x=0, y=1.0, z=1) + off,
        test_lw2["A12"].from_center_cartesian(x=0, y=-1.0, z=1) + off,
    ]
    # Test that labware in slot 6 results in modified edge list
    res2 = build_edges(
        test_lw2["A12"],
        1.0,
        Mount.LEFT,
        ctx._core.get_deck(),  # type: ignore[attr-defined]
        version=APIVersion(2, 4),
    )
    assert res2 == left_pip_edges


# TODO(mm, 2023-04-28): See note on test_build_edges_left_pipette().
@pytest.mark.apiv2_non_pe_only
def test_build_edges_right_pipette(ctx: ProtocolContext) -> None:
    test_lw = ctx.load_labware("corning_96_wellplate_360ul_flat", "2")
    test_lw2 = ctx.load_labware("corning_96_wellplate_360ul_flat", "6")
    mod = ctx.load_module("magnetic module", "1")
    mod.load_labware("corning_96_wellplate_360ul_flat")
    off = Point(0, 0, 1.0)
    right_pip_edges = [
        test_lw["A1"].from_center_cartesian(x=1.0, y=0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=1.0, z=1) + off,
        test_lw["A1"].from_center_cartesian(x=0, y=-1.0, z=1) + off,
    ]
    # Test that module in slot 1 results in modified edge list
    res = build_edges(
        test_lw["A1"],
        1.0,
        Mount.RIGHT,
        ctx._core.get_deck(),  # type: ignore[attr-defined]
        version=APIVersion(2, 4),
    )
    assert res == right_pip_edges

    right_pip_edges = [
        test_lw2["A12"].from_center_cartesian(x=1.0, y=0, z=1) + off,
        test_lw2["A12"].from_center_cartesian(x=-1.0, y=0, z=1) + off,
        test_lw2["A12"].from_center_cartesian(x=0, y=0, z=1) + off,
        test_lw2["A12"].from_center_cartesian(x=0, y=1.0, z=1) + off,
        test_lw2["A12"].from_center_cartesian(x=0, y=-1.0, z=1) + off,
    ]
    # Test that labware in slot 6 results in unmodified edge list
    res2 = build_edges(
        test_lw2["A12"],
        1.0,
        Mount.RIGHT,
        ctx._core.get_deck(),  # type: ignore[attr-defined]
        version=APIVersion(2, 4),
    )
    assert res2 == right_pip_edges


@pytest.mark.parametrize(
    "data,level,desired",
    [
        ({"2.0": 5}, APIVersion(2, 0), 5),
        ({"2.0": 5}, APIVersion(2, 5), 5),
        ({"2.6": 4, "2.0": 5}, APIVersion(2, 1), 5),
        ({"2.6": 4, "2.0": 5}, APIVersion(2, 6), 4),
        ({"2.0": 5, "2.6": 4}, APIVersion(2, 3), 5),
        ({"2.0": 5, "2.6": 4}, APIVersion(2, 6), 4),
    ],
)
def test_find_value_for_api_version(
    data: Dict[str, float], level: APIVersion, desired: float
) -> None:
    assert find_value_for_api_version(level, data) == desired



from unittest.mock import MagicMock

import pytest

from opentrons_shared_data.labware.types import LabwareDefinition2

from opentrons.hardware_control.modules.types import TemperatureModuleModel
from opentrons.protocol_api import OFF_DECK, labware
from opentrons.protocol_api.core.legacy import module_geometry
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.module_geometry import ModuleGeometry
from opentrons.protocol_api.labware import Labware
from opentrons.protocols.api_support.deck_type import STANDARD_OT2_DECK
from opentrons.protocols.api_support.labware_like import LabwareLike, LabwareLikeType
from opentrons.types import Location, Point


@pytest.fixture(scope="session")
def trough_definition() -> LabwareDefinition2:
    result = labware.get_labware_definition("usascientific_12_reservoir_22ml")
    assert result["schemaVersion"] == 2  # For type checking.
    return result


@pytest.fixture(scope="session")
def trough(trough_definition: LabwareDefinition2) -> Labware:
    deck = Deck(deck_type=STANDARD_OT2_DECK)
    return labware.load_from_definition(trough_definition, deck.position_for(1))


@pytest.fixture(scope="session")
def module() -> ModuleGeometry:
    deck = Deck(deck_type=STANDARD_OT2_DECK)
    mod = module_geometry.create_geometry(
        definition=module_geometry.load_definition(
            TemperatureModuleModel.TEMPERATURE_V2
        ),
        parent=deck.position_for("6"),
        configuration=None,
    )
    return mod


@pytest.fixture(scope="session")
def mod_trough(
    trough_definition: LabwareDefinition2, module: ModuleGeometry
) -> Labware:
    mod_trough = module.add_labware(
        labware.load_from_definition(trough_definition, module.location)
    )
    return mod_trough


def test_labware(trough: Labware) -> None:
    ll = LabwareLike(trough)
    assert ll.has_parent is True
    assert ll.parent.object == trough.parent
    assert ll.object == trough
    assert ll.object_type == LabwareLikeType.LABWARE


def test_well(trough: Labware) -> None:
    well = trough["A1"]
    ll = LabwareLike(well)
    assert ll.has_parent is True
    assert ll.parent.object == trough
    assert ll.object == well
    assert ll.object_type == LabwareLikeType.WELL


def test_module(module: ModuleGeometry) -> None:
    ll = LabwareLike(module)
    assert ll.has_parent is True
    assert ll.parent.object is module.parent
    assert ll.object is module
    assert ll.object_type == LabwareLikeType.MODULE
    assert ll.is_module
    assert ll.as_module() == module


def test_slot() -> None:
    ll = LabwareLike("1")
    assert ll.has_parent is False
    assert ll.parent.object is None
    assert ll.object == "1"
    assert ll.object_type == LabwareLikeType.SLOT


def test_empty() -> None:
    ll = LabwareLike(None)
    assert ll.has_parent is False
    assert ll.parent.object is None
    assert ll.object is None
    assert ll.object_type == LabwareLikeType.NONE


def test_off_deck() -> None:
    ll = LabwareLike(OFF_DECK)
    assert ll.has_parent is False
    assert ll.parent.object is None
    assert ll.object is OFF_DECK
    assert ll.object_type == LabwareLikeType.OFF_DECK


def test_module_parent(
    trough: Labware, module: ModuleGeometry, mod_trough: Labware
) -> None:
    assert LabwareLike(mod_trough).module_parent() == module
    assert LabwareLike(mod_trough["A1"]).module_parent() == module
    assert LabwareLike(module).module_parent() == module
    assert LabwareLike(trough).module_parent() is None
    assert LabwareLike("1").module_parent() is None


def test_first_parent(
    trough: Labware, module: ModuleGeometry, mod_trough: Labware
) -> None:
    assert LabwareLike(trough).first_parent() == "1"
    assert LabwareLike(trough["A2"]).first_parent() == "1"
    assert LabwareLike(None).first_parent() is None
    assert LabwareLike("6").first_parent() == "6"

    assert LabwareLike(mod_trough["A5"]).first_parent() == "6"
    assert LabwareLike(mod_trough).first_parent() == "6"
    assert LabwareLike(module).first_parent() == "6"

    # Set up recursion cycle test.
    mock_labware_geometry = MagicMock()
    mock_labware_geometry.parent = Location(
        point=Point(x=-0.75, y=-0.75, z=1), labware=mod_trough
    )
    assert hasattr(mod_trough._core, "get_geometry")
    mod_trough._core.get_geometry = MagicMock(return_value=mock_labware_geometry)

    with pytest.raises(RuntimeError):
        # make sure we catch cycles
        LabwareLike(mod_trough).first_parent()



"""Test Temperature Module's set target temperature command implementation."""

from decoy import Decoy, matchers

from opentrons.hardware_control.modules import TempDeck
from opentrons.protocol_engine.actions import Action, ActionDispatcher, StartTaskAction
from opentrons.protocol_engine.commands import temperature_module
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.temperature_module.set_target_temperature import (
    SetTargetTemperatureImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler, TaskHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    TemperatureModuleId,
    TemperatureModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types.tasks import Task


async def test_set_target_temperature(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    real_task_handler: TaskHandler,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
) -> None:
    """It should be able to set the specified module's target temperature."""
    subject = SetTargetTemperatureImpl(
        state_view=state_view, equipment=equipment, task_handler=real_task_handler
    )

    data = temperature_module.SetTargetTemperatureParams(
        moduleId="tempdeck-id", celsius=1.23, taskId="taskId"
    )

    module_substate = decoy.mock(cls=TemperatureModuleSubState)
    tempdeck_hardware = decoy.mock(cls=TempDeck)

    decoy.when(
        state_view.modules.get_temperature_module_substate(module_id="tempdeck-id")
    ).then_return(module_substate)
    decoy.when(model_utils.ensure_id("taskId")).then_return("taskId")
    decoy.when(module_substate.module_id).then_return(
        TemperatureModuleId("tempdeck-id")
    )

    # Stub temperature validation
    decoy.when(module_substate.validate_target_temperature(celsius=1.23)).then_return(1)

    # Get stubbed hardware module
    decoy.when(
        equipment.get_module_hardware_api(TemperatureModuleId("tempdeck-id"))
    ).then_return(tempdeck_hardware)

    task: Task | None = None

    def _capture_task(action: Action) -> None:
        nonlocal task
        assert isinstance(action, StartTaskAction)
        task = action.task

    decoy.when(
        action_dispatcher.dispatch(StartTaskAction(task=matchers.Anything()))  # type: ignore[func-returns-value]
    ).then_do(_capture_task)

    result = await subject.execute(data)
    assert task is not None
    await task.asyncioTask
    decoy.verify(
        await tempdeck_hardware.start_set_temperature(celsius=1),
        times=1,
    )
    decoy.verify(await tempdeck_hardware.await_temperature(awaiting_temperature=1))
    assert result == SuccessData(
        public=temperature_module.SetTargetTemperatureResult(
            targetTemperature=1, taskId="taskId"
        ),
    )



"""Tests for transfer_liquid_utils."""

from contextlib import nullcontext as does_not_raise
from logging import Logger
from typing import Any, ContextManager

import pytest
from decoy import Decoy

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    Coordinate,
    PositionReference,
)
from opentrons_shared_data.pipette.pipette_definition import ValidNozzleMaps
from tests.opentrons.protocol_engine.pipette_fixtures import (
    EIGHT_CHANNEL_COLS,
    EIGHT_CHANNEL_MAP,
    EIGHT_CHANNEL_ROWS,
    NINETY_SIX_COLS,
    NINETY_SIX_MAP,
    NINETY_SIX_ROWS,
)

from .labware_well_fixtures import WELLS_BY_COLUMN_96, WELLS_BY_COLUMN_384
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_api._liquid_properties import TipPosition
from opentrons.protocol_api.core.engine import WellCore
from opentrons.protocol_api.disposal_locations import (
    _TRASH_BIN_CUTOUT_FIXTURE,
    DisposalOffset,
    TrashBin,
)
from opentrons.protocol_api.labware import Labware, Well
from opentrons.protocol_engine import ProtocolEngineError
from opentrons.protocol_engine.clients.sync_client import SyncClient as EngineClient
from opentrons.protocol_engine.errors import (
    IncompleteLabwareDefinitionError,
    LiquidHeightUnknownError,
)
from opentrons.protocol_engine.types.liquid_level_detection import (
    LiquidTrackingType,
)
from opentrons.protocols.advanced_control.transfers.transfer_liquid_utils import (
    LocationCheckDescriptors,
    get_blowout_location_for_trash,
    group_wells_for_multi_channel_transfer,
    raise_if_location_inside_liquid,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import DeckSlotName, Location, Point

_96_FULL_MAP = NozzleMap.build(
    physical_nozzles=NINETY_SIX_MAP,
    physical_rows=NINETY_SIX_ROWS,
    physical_columns=NINETY_SIX_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="H12",
    valid_nozzle_maps=ValidNozzleMaps(
        maps={
            "Full": sum(
                [
                    NINETY_SIX_ROWS["A"],
                    NINETY_SIX_ROWS["B"],
                    NINETY_SIX_ROWS["C"],
                    NINETY_SIX_ROWS["D"],
                    NINETY_SIX_ROWS["E"],
                    NINETY_SIX_ROWS["F"],
                    NINETY_SIX_ROWS["G"],
                    NINETY_SIX_ROWS["H"],
                ],
                [],
            )
        }
    ),
)

_96_COL1_MAP = NozzleMap.build(
    physical_nozzles=NINETY_SIX_MAP,
    physical_rows=NINETY_SIX_ROWS,
    physical_columns=NINETY_SIX_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="H1",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Column1": NINETY_SIX_COLS["1"]}),
)

_96_COL12_MAP = NozzleMap.build(
    physical_nozzles=NINETY_SIX_MAP,
    physical_rows=NINETY_SIX_ROWS,
    physical_columns=NINETY_SIX_COLS,
    starting_nozzle="A12",
    back_left_nozzle="A12",
    front_right_nozzle="H12",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Column12": NINETY_SIX_COLS["12"]}),
)

_96_ROW_A_MAP = NozzleMap.build(
    physical_nozzles=NINETY_SIX_MAP,
    physical_rows=NINETY_SIX_ROWS,
    physical_columns=NINETY_SIX_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="A12",
    valid_nozzle_maps=ValidNozzleMaps(maps={"RowA": NINETY_SIX_ROWS["A"]}),
)

_96_ROW_H_MAP = NozzleMap.build(
    physical_nozzles=NINETY_SIX_MAP,
    physical_rows=NINETY_SIX_ROWS,
    physical_columns=NINETY_SIX_COLS,
    starting_nozzle="H1",
    back_left_nozzle="H1",
    front_right_nozzle="H12",
    valid_nozzle_maps=ValidNozzleMaps(maps={"RowH": NINETY_SIX_ROWS["H"]}),
)

_8_FULL_MAP = NozzleMap.build(
    physical_nozzles=EIGHT_CHANNEL_MAP,
    physical_rows=EIGHT_CHANNEL_ROWS,
    physical_columns=EIGHT_CHANNEL_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="H1",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Full": EIGHT_CHANNEL_COLS["1"]}),
)


@pytest.fixture
def mock_96_well_labware(decoy: Decoy) -> Labware:
    """Get a mock 96 well labware."""
    mock_96_well_labware = decoy.mock(cls=Labware)
    decoy.when(mock_96_well_labware.parameters).then_return({"format": "96Standard"})  # type: ignore[typeddict-item]
    labware_wells_by_column = []
    for column in WELLS_BY_COLUMN_96:
        wells_by_column = []
        for well_name in column:
            mock_well = decoy.mock(cls=Well)
            decoy.when(mock_well.well_name).then_return(well_name)
            wells_by_column.append(mock_well)
        labware_wells_by_column.append(wells_by_column)
    decoy.when(mock_96_well_labware.columns()).then_return(labware_wells_by_column)
    return mock_96_well_labware


@pytest.fixture
def mock_384_well_labware(decoy: Decoy) -> Labware:
    """Get a mock 96 well labware."""
    mock_384_well_labware = decoy.mock(cls=Labware)
    decoy.when(mock_384_well_labware.parameters).then_return({"format": "384Standard"})  # type: ignore[typeddict-item]
    labware_wells_by_column = []
    for column in WELLS_BY_COLUMN_384:
        wells_by_column = []
        for well_name in column:
            mock_well = decoy.mock(cls=Well)
            decoy.when(mock_well.well_name).then_return(well_name)
            wells_by_column.append(mock_well)
        labware_wells_by_column.append(wells_by_column)
    decoy.when(mock_384_well_labware.columns()).then_return(labware_wells_by_column)
    return mock_384_well_labware


@pytest.mark.parametrize(
    argnames=["liquid_height", "pip_location", "well_bottom", "expected_raise"],
    argvalues=[
        (
            1.0,
            Location(point=Point(4, 5, 6), labware=None),
            Point(0, 0, 0),
            does_not_raise(),
        ),
        (
            100.0,
            Location(point=Point(4, 5, 6), labware=None),
            Point(0, 0, 0),
            pytest.raises(
                RuntimeError,
                match="Retract end location Location\\(point=Point\\(x=4, y=5, z=6\\), labware=,"
                " meniscus_tracking=None\\) is inside the liquid in well Well A1 of"
                " test_labware when it should be outside\\(above\\) the liquid.",
            ),
        ),
        (
            1,
            Location(point=Point(5, 6, 7), labware=None),
            Point(10, 11, 12),
            pytest.raises(
                RuntimeError,
                match="Retract end location Location\\(point=Point\\(x=5, y=6, z=7\\), labware=,"
                " meniscus_tracking=None\\) is inside the liquid in well Well A1 of"
                " test_labware when it should be outside\\(above\\) the liquid.",
            ),
        ),
        (
            None,
            Location(point=Point(5, 6, 7), labware=None),
            Point(10, 11, 12),
            does_not_raise(),
        ),
    ],
)
def test_raise_only_if_pip_location_inside_liquid(
    decoy: Decoy,
    liquid_height: LiquidTrackingType,
    pip_location: Location,
    well_bottom: Point,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise an error if we have access to liquid height and pipette is in liquid."""
    well_core = decoy.mock(cls=WellCore)
    location_descriptors = LocationCheckDescriptors(
        location_type="retract end",
        pipetting_action="aspirate",
    )
    logger = decoy.mock(cls=Logger)

    decoy.when(well_core.current_liquid_height()).then_return(liquid_height)
    decoy.when(well_core.get_bottom(0)).then_return(well_bottom)
    decoy.when(well_core.get_display_name()).then_return("Well A1 of test_labware")
    with expected_raise:
        raise_if_location_inside_liquid(
            location=pip_location,
            well_core=well_core,
            location_check_descriptors=location_descriptors,
            logger=logger,
        )


@pytest.mark.parametrize(
    "error_raised", [LiquidHeightUnknownError(), IncompleteLabwareDefinitionError()]
)
def test_log_warning_if_pip_location_cannot_be_validated(
    decoy: Decoy,
    error_raised: ProtocolEngineError,
) -> None:
    """It should log a warning if we don't have access to liquid height."""
    pip_location = Location(point=Point(1, 2, 3), labware=None)
    well_core = decoy.mock(cls=WellCore)
    location_descriptors = LocationCheckDescriptors(
        location_type="retract end",
        pipetting_action="aspirate",
    )
    logger = decoy.mock(cls=Logger)

    decoy.when(well_core.current_liquid_height()).then_raise(error_raised)
    decoy.when(well_core.get_bottom(0)).then_return(Point(0, 0, 0))
    decoy.when(well_core.get_display_name()).then_return("Well A1 of test_labware")
    raise_if_location_inside_liquid(
        location=pip_location,
        well_core=well_core,
        location_check_descriptors=location_descriptors,
        logger=logger,
    )
    decoy.verify(
        logger.info(
            "Could not verify height of liquid in well Well A1 of test_labware, either"
            " because the liquid in this well has not been probed or"
            " liquid was not loaded in this well using `load_liquid` or"
            " inner geometry is not available for the target well."
            " Proceeding without verifying if retract end location is outside the liquid."
        )
    )


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_8_FULL_MAP, _96_COL1_MAP, _96_COL12_MAP]
)
def test_grouping_wells_for_column_96_plate(
    nozzle_map: NozzleMap, mock_96_well_labware: Labware, decoy: Decoy
) -> None:
    """It should group two columns into A1 and A2."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(16)]
    for mock_well, well_name in zip(
        mock_wells, WELLS_BY_COLUMN_96[0] + WELLS_BY_COLUMN_96[1]
    ):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_96_well_labware)

    wells = group_wells_for_multi_channel_transfer(mock_wells, nozzle_map, "source")
    assert len(wells) == 2
    assert wells[0].well_name == "A1"
    assert wells[1].well_name == "A2"


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_8_FULL_MAP, _96_COL1_MAP, _96_COL12_MAP]
)
def test_grouping_wells_for_column_384_plate(
    nozzle_map: NozzleMap, mock_384_well_labware: Labware, decoy: Decoy
) -> None:
    """It should group two columns into A1, B1, A2 and B2."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(32)]
    for mock_well, well_name in zip(
        mock_wells, WELLS_BY_COLUMN_384[0] + WELLS_BY_COLUMN_384[1]
    ):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_384_well_labware)

    wells = group_wells_for_multi_channel_transfer(mock_wells, nozzle_map, "source")
    assert len(wells) == 4
    assert wells[0].well_name == "A1"
    assert wells[1].well_name == "B1"
    assert wells[2].well_name == "A2"
    assert wells[3].well_name == "B2"


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_8_FULL_MAP, _96_COL1_MAP, _96_COL12_MAP]
)
def test_grouping_wells_for_column_96_plate_raises(
    nozzle_map: NozzleMap, mock_96_well_labware: Labware, decoy: Decoy
) -> None:
    """It should raise if a valid grouping can't be found for all wells."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(18)]
    for mock_well, well_name in zip(
        mock_wells, WELLS_BY_COLUMN_96[0] + WELLS_BY_COLUMN_96[1] + ["A3", "B3"]
    ):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_96_well_labware)

    # leftover wells
    with pytest.raises(ValueError, match="Pipette will access source wells"):
        group_wells_for_multi_channel_transfer(mock_wells, nozzle_map, "source")

    # non-contiguous wells from the same labware
    with pytest.raises(ValueError, match="Could not group source wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:7] + [mock_wells[-1], mock_wells[7]], nozzle_map, "source"
        )

    other_labware = decoy.mock(cls=Labware)
    decoy.when(other_labware.parameters).then_return({"format": "96Standard"})  # type: ignore[typeddict-item]
    other_well = decoy.mock(cls=Well)
    decoy.when(other_well.well_name).then_return("H1")
    decoy.when(other_well.parent).then_return(other_labware)

    # non-contiguous wells from different labware, well name is correct though
    with pytest.raises(ValueError, match="Could not group source wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:7] + [other_well], nozzle_map, "source"
        )


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_8_FULL_MAP, _96_COL1_MAP, _96_COL12_MAP]
)
def test_grouping_wells_for_column_384_plate_raises(
    nozzle_map: NozzleMap, mock_384_well_labware: Labware, decoy: Decoy
) -> None:
    """It should raise if a valid grouping can't be found for all wells."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(32)]
    for mock_well, well_name in zip(
        mock_wells, WELLS_BY_COLUMN_384[0] + WELLS_BY_COLUMN_384[1]
    ):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_384_well_labware)

    # leftover wells
    with pytest.raises(ValueError, match="Pipette will access destination wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:-1], nozzle_map, "destination"
        )

    # non-contiguous or every other wells from the same labware
    with pytest.raises(ValueError, match="Could not group"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:2] + [mock_wells[-1]], nozzle_map, "source"
        )
        group_wells_for_multi_channel_transfer(
            mock_wells[:-1] + [mock_wells[0]], nozzle_map, "destination"
        )

    other_labware = decoy.mock(cls=Labware)
    decoy.when(other_labware.parameters).then_return({"format": "384Standard"})  # type: ignore[typeddict-item]
    other_well = decoy.mock(cls=Well)
    decoy.when(other_well.well_name).then_return("P1")
    decoy.when(other_well.parent).then_return(other_labware)

    # non-contiguous wells from different labware, well name is correct though
    with pytest.raises(ValueError, match="Could not group source wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:15] + [other_well], nozzle_map, "source"
        )


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_96_ROW_A_MAP, _96_ROW_H_MAP]
)
def test_grouping_wells_for_row_96_plate(
    nozzle_map: NozzleMap, mock_96_well_labware: Labware, decoy: Decoy
) -> None:
    """It should group two rows into A1 and B1."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(24)]
    first_two_row_well_names = [f"A{i}" for i in range(1, 13)] + [
        f"B{i}" for i in range(1, 13)
    ]
    for mock_well, well_name in zip(mock_wells, first_two_row_well_names):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_96_well_labware)

    wells = group_wells_for_multi_channel_transfer(
        mock_wells, nozzle_map, "destination"
    )
    assert len(wells) == 2
    assert wells[0].well_name == "A1"
    assert wells[1].well_name == "B1"


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_96_ROW_A_MAP, _96_ROW_H_MAP]
)
def test_grouping_wells_for_row_384_plate(
    nozzle_map: NozzleMap, mock_384_well_labware: Labware, decoy: Decoy
) -> None:
    """It should group two columns into A1, A2, B1, B2."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(48)]
    first_two_row_well_names = [f"A{i}" for i in range(1, 25)] + [
        f"B{i}" for i in range(1, 25)
    ]
    for mock_well, well_name in zip(mock_wells, first_two_row_well_names):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_384_well_labware)

    wells = group_wells_for_multi_channel_transfer(mock_wells, nozzle_map, "source")
    assert len(wells) == 4
    assert wells[0].well_name == "A1"
    assert wells[1].well_name == "A2"
    assert wells[2].well_name == "B1"
    assert wells[3].well_name == "B2"


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_96_ROW_A_MAP, _96_ROW_H_MAP]
)
def test_grouping_wells_for_row_96_plate_raises(
    nozzle_map: NozzleMap, mock_96_well_labware: Labware, decoy: Decoy
) -> None:
    """It should raise if a valid grouping can't be found for all wells."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(24)]
    first_two_row_well_names = [f"A{i}" for i in range(1, 13)] + [
        f"B{i}" for i in range(1, 13)
    ]
    for mock_well, well_name in zip(mock_wells, first_two_row_well_names):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_96_well_labware)

    # leftover wells
    with pytest.raises(ValueError, match="Pipette will access source wells"):
        group_wells_for_multi_channel_transfer(mock_wells[:-1], nozzle_map, "source")

    # non-contiguous wells from the same labware
    with pytest.raises(ValueError, match="Could not group source wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:11] + [mock_wells[-1], mock_wells[11]], nozzle_map, "source"
        )

    other_labware = decoy.mock(cls=Labware)
    decoy.when(other_labware.parameters).then_return({"format": "96Standard"})  # type: ignore[typeddict-item]
    other_well = decoy.mock(cls=Well)
    decoy.when(other_well.well_name).then_return("A12")
    decoy.when(other_well.parent).then_return(other_labware)

    # non-contiguous wells from different labware, well name is correct though
    with pytest.raises(ValueError, match="Could not group destination wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:11] + [other_well], nozzle_map, "destination"
        )


@pytest.mark.parametrize(
    argnames="nozzle_map", argvalues=[_96_ROW_A_MAP, _96_ROW_H_MAP]
)
def test_grouping_wells_for_row_384_plate_raises(
    nozzle_map: NozzleMap, mock_384_well_labware: Labware, decoy: Decoy
) -> None:
    """It should raise if a valid grouping can't be found for all wells."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(48)]
    first_two_row_well_names = [f"A{i}" for i in range(1, 25)] + [
        f"B{i}" for i in range(1, 25)
    ]
    for mock_well, well_name in zip(mock_wells, first_two_row_well_names):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_384_well_labware)

    # leftover wells
    with pytest.raises(ValueError, match="Pipette will access destination wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:-1], nozzle_map, "destination"
        )

    # non-contiguous or every other wells from the same labware
    with pytest.raises(ValueError, match="Could not group"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:2] + [mock_wells[-1]], nozzle_map, "destination"
        )
        group_wells_for_multi_channel_transfer(
            mock_wells[:-1] + [mock_wells[0]], nozzle_map, "source"
        )

    other_labware = decoy.mock(cls=Labware)
    decoy.when(other_labware.parameters).then_return({"format": "384Standard"})  # type: ignore[typeddict-item]
    other_well = decoy.mock(cls=Well)
    decoy.when(other_well.well_name).then_return("A24")
    decoy.when(other_well.parent).then_return(other_labware)

    # non-contiguous wells from different labware, well name is correct though
    with pytest.raises(ValueError, match="Could not group destination wells"):
        group_wells_for_multi_channel_transfer(
            mock_wells[:23] + [other_well], nozzle_map, "destination"
        )


def test_grouping_wells_for_full_96_plate(
    mock_96_well_labware: Labware, decoy: Decoy
) -> None:
    """It should group a whole 96 well plate into A1."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(96)]
    for mock_well, well_name in zip(mock_wells, NINETY_SIX_MAP.keys()):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_96_well_labware)

    wells = group_wells_for_multi_channel_transfer(
        mock_wells, _96_FULL_MAP, "destination"
    )
    assert len(wells) == 1
    assert wells[0].well_name == "A1"


def test_grouping_wells_for_full_384_plate(
    mock_384_well_labware: Labware, decoy: Decoy
) -> None:
    """It should a whole 384 well plate into A1, B1, A2 and B2."""
    mock_wells = [decoy.mock(cls=Well) for _ in range(384)]
    flat_384_well_names = [well for column in WELLS_BY_COLUMN_384 for well in column]
    for mock_well, well_name in zip(mock_wells, flat_384_well_names):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_384_well_labware)

    wells = group_wells_for_multi_channel_transfer(mock_wells, _96_FULL_MAP, "source")
    assert len(wells) == 4
    assert wells[0].well_name == "A1"
    assert wells[1].well_name == "B1"
    assert wells[2].well_name == "A2"
    assert wells[3].well_name == "B2"


def test_grouping_wells_raises_for_unsupported_configuration() -> None:
    """It should raise if the well configuration is not supported."""
    nozzle_map = NozzleMap.build(
        physical_nozzles=EIGHT_CHANNEL_MAP,
        physical_rows=EIGHT_CHANNEL_ROWS,
        physical_columns=EIGHT_CHANNEL_COLS,
        starting_nozzle="A1",
        back_left_nozzle="A1",
        front_right_nozzle="D1",
        valid_nozzle_maps=ValidNozzleMaps(maps={"Half": ["A1", "B1", "C1", "D1"]}),
    )
    with pytest.raises(ValueError, match="Unsupported nozzle configuration"):
        group_wells_for_multi_channel_transfer([], nozzle_map, "source")


@pytest.mark.parametrize(
    argnames="nozzle_map",
    argvalues=[
        _8_FULL_MAP,
        _96_FULL_MAP,
        _96_COL1_MAP,
        _96_COL12_MAP,
        _96_ROW_A_MAP,
        _96_ROW_H_MAP,
    ],
)
def test_grouping_well_returns_all_wells_for_non_96_or_384_plate(
    nozzle_map: NozzleMap, decoy: Decoy
) -> None:
    """It should return all wells if parent labware is not a 96 or 384 well plate"""
    mock_reservoir = decoy.mock(cls=Labware)
    decoy.when(mock_reservoir.parameters).then_return({"format": "reservoir"})  # type: ignore[typeddict-item]

    mock_wells = [decoy.mock(cls=Well) for _ in range(12)]
    for mock_well, well_name in zip(mock_wells, [f"A{i}" for i in range(1, 13)]):
        decoy.when(mock_well.well_name).then_return(well_name)
        decoy.when(mock_well.parent).then_return(mock_reservoir)

    result = group_wells_for_multi_channel_transfer(mock_wells, nozzle_map, "source")
    assert result == mock_wells


@pytest.mark.parametrize(
    argnames="target_tip_position, expected_offset_from_top",
    argvalues=[
        (
            TipPosition(
                _position_reference=PositionReference.WELL_TOP,
                _offset=Coordinate(x=0, y=0, z=0),
            ),
            DisposalOffset(x=0, y=0, z=0),
        ),
        (
            TipPosition(
                _position_reference=PositionReference.WELL_BOTTOM,
                _offset=Coordinate(x=0, y=0, z=0),
            ),
            DisposalOffset(x=0, y=0, z=-10),
        ),
        (
            TipPosition(
                _position_reference=PositionReference.WELL_CENTER,
                _offset=Coordinate(x=1, y=2, z=3),
            ),
            DisposalOffset(x=1, y=2, z=-2),
        ),
    ],
)
def test_blowout_location_for_trash(
    decoy: Decoy,
    target_tip_position: TipPosition,
    expected_offset_from_top: DisposalOffset,
) -> None:
    """Should return the expected blowout location for all given disposal locations."""
    engine_client = decoy.mock(cls=EngineClient)
    trash = TrashBin(
        location=DeckSlotName.SLOT_A1,
        addressable_area_name="moveableTrashD3",
        api_version=APIVersion(2, 28),
        engine_client=engine_client,
    )
    decoy.when(
        engine_client.state.addressable_areas.get_fixture_height(
            _TRASH_BIN_CUTOUT_FIXTURE
        )
    ).then_return(10)
    assert (
        get_blowout_location_for_trash(trash, target_tip_position).offset
        == expected_offset_from_top
    )


def test_blowout_location_for_trash_raises_when_position_reference_is_liquid_meniscus(
    decoy: Decoy,
) -> None:
    """Should raise ValueError when position reference is PositionReference.LIQUID_MENISCUS."""
    with pytest.raises(ValueError):
        get_blowout_location_for_trash(
            decoy.mock(cls=TrashBin),
            TipPosition(
                _position_reference=PositionReference.LIQUID_MENISCUS,
                _offset=Coordinate(x=0, y=0, z=0),
            ),
        )
